#include "swuart.h"


AsynchronousStates_t state;     //!< Holds the state of the UART.
uint8_t SwUartTXData;     //!< Data to be transmitted.
uint8_t SwUartTXBitCount; //!< TX bit counter.
uint8_t SwUartRXData;     //!< Storage for received bits.
uint8_t SwUartRXBitCount; //!< RX bit counter.


void swuart_init() {
  TRXPORT |= ( 1 << RX_PIN );       // RX_PIN is input, tri-stated.
  TRXDDR |= ( 1 << TX_PIN );        // TX_PIN is output.
  SET_TX_PIN( );                    // Set the TX line to idle state.

  // Timer0
  DISABLE_TIMER_INTERRUPT( );
  TCCR = 0x00;    //Init.
  TCCR_P = 0x00;    //Init.
  TCCR |= (1 << WGM01);			// Timer in CTC mode.
  TCCR_P |=  ( 1 << CS01 );        // Divide by 8 prescaler.


  //External interrupt
  EXT_ICR = 0x00;                   // Init.
  PCMSK |= ( 1 << RX_PCINT );		// Pin change interrupt number
  CLEAR_EXTERNAL0_INTERRUPT( );
  ENABLE_EXTERNAL0_INTERRUPT( );    // Turn external interrupt on.

  //Internal State Variable
  state = IDLE;

}
 
/*
 * sendet ein Byte ber den SW-UART
 */
void swuart_putchar(uint8_t c) {
  while( state != IDLE ){}			// warte bis gesendet werden darf

  state = TRANSMIT_START_BIT;
  DISABLE_EXTERNAL0_INTERRUPT( );   // Disable reception.
  SwUartTXData = c;             	// Put byte into TX buffer.
  SwUartTXBitCount = 0;         
  

  TCCR_P &= ~( 1 << CS01 );         // Reset prescaler counter.
  TCNT0 = 0;                        // Clear counter register.
  OCR = TICKS2WAITONE;
  TCCR_P |= ( 1 << CS01 );          // CTC mode. Start prescaler clock.

  ENABLE_TIMER_INTERRUPT( );        // Enable interrupt
}

/*
 * sendet einen String ber den SW-UART
 */
void swuart_print( const unsigned char *data ) {
  while( *data != '\0' ) {
    swuart_putchar( *data++ );
  }
}

/*
 * Liest ein Byte vom SW-UART
 * Es wird gewartet bis ein Byte kommt
 */
uint8_t swuart_getchar() {

	while (state != DATA_PENDING) {} // Warte bis du Daten hast
	state = IDLE;					 // Setze den Status auf IDLE
	return SwUartRXData;			 // und liefere die empfangenen Daten zurck
}

/*
 * Lies den aktuellen UART-Status aus
 * Data Polling via if (swuart_getstate() == DATA_PENDING)
 */
AsynchronousStates_t swuart_getstate () {
	return state;
}


/*
 * Timer-Interrupt Routine
 * State-Machine ber den AsynchronousStates_t
 */
ISR(TIMER0_COMPA_vect) {
  switch (state) {
  // Transmit the start bit
  case TRANSMIT_START_BIT:
	CLEAR_TX_PIN();               // Clear TX line...start of preamble.
  	state = TRANSMIT;
	break;
  // Transmit Byte.
  case TRANSMIT:
    // Output the TX buffer.
    if( SwUartTXBitCount < 8 ) {            
      if( SwUartTXData & 0x01 ) {           // If the LSB of the TX buffer is 1:
        SET_TX_PIN();                       // Send a logic 1 on the TX_PIN.
      }
      else {                                // Otherwise:
        CLEAR_TX_PIN();                     // Send a logic 0 on the TX_PIN.
      }
      SwUartTXData = SwUartTXData >> 1;     // Bitshift the TX buffer and
      SwUartTXBitCount++;                   // increment TX bit counter.
    }

    //Send stop bit.
    else {
      state = TRANSMIT_STOP_BIT1;
    }


  break;

  // Go to idle after stop bit was sent.
  case TRANSMIT_STOP_BIT1: 
    SET_TX_PIN();                         // Output a logic 1.
	state = TRANSMIT_STOP_BIT2;
	break;

  // Go to idle after stop bit was sent.
  case TRANSMIT_STOP_BIT2: 
    SET_TX_PIN();                         // Output a logic 1.
	state = TRANSMIT_END;
	break;

  case TRANSMIT_END:
    DISABLE_TIMER_INTERRUPT( );           // Stop the timer interrupts.
    state = IDLE;                         // Go back to idle.
    ENABLE_EXTERNAL0_INTERRUPT( );        // Enable reception again.
  break;

  //Receive Byte.
  case RECEIVE:
	OCR = TICKS2WAITONE - 10;              // Count one period after the falling edge is trigged.


    //Receiving, LSB first.
    if( SwUartRXBitCount < 8 ) {
        SwUartRXBitCount++;
        SwUartRXData = (SwUartRXData>>1);   // Shift due to receiving LSB first.
        if( GET_RX_PIN( ) != 0 ) {
            SwUartRXData |= 0x80;           // If a logical 1 is read, let the data mirror this.
        }
    }

    //Done receiving
    else {
        state = DATA_PENDING;               // Enter DATA_PENDING when one byte is received.
        DISABLE_TIMER_INTERRUPT( );         // Disable this interrupt.
        EXT_IFR = (1 << PCIF );             // Reset flag not to enter the ISR one extra time.
        ENABLE_EXTERNAL0_INTERRUPT( );      // Enable interrupt to receive more bytes.
    }
  break;

  // Unknown state.
  default:        
    state = IDLE;                           // Error, should not occur. Going to a safe state.
  }
}

/*
 * Externer Interrupt- Es wird etwas empfangen..
 */
ISR (PCINT0_vect) {
  //if ((TRXPIN & (1 << RX_PIN)) == 0 ) { // beim Flankenwechsel nach low
	  state = RECEIVE;                  // Change state
	  DISABLE_EXTERNAL0_INTERRUPT( );   // Disable interrupt during the data bits.

	  DISABLE_TIMER_INTERRUPT( );       // Disable timer to change its registers.
	  TCCR_P &= ~( 1 << CS01 );         // Reset prescaler counter.

	  TCNT0 = INTERRUPT_EXEC_CYCL;      // Clear counter register. Include time to run interrupt rutine.

	  OCR = TICKS2WAITONE_HALF;         // Count one and a half period into the future.

	  TCCR_P |= ( 1 << CS01 );          // Start prescaler clock.

	  SwUartRXBitCount = 0;             // Clear received bit counter.
	  SwUartRXData = 0;
	  CLEAR_TIMER_INTERRUPT( );         // Clear interrupt bits
	  ENABLE_TIMER_INTERRUPT( );        // Enable timer0 interrupt on again
  //}
}


